/*
This code is copyright (C) 1998-1999 Robert O'Callahan.
See LICENSE.TXT for the license.
*/

#include "ttxssh.h"
#include "util.h"

#define READAMOUNT 60000

void PKT_init(PTInstVar pvar) {
  buf_create(&pvar->pkt_state.buf, &pvar->pkt_state.buflen);
  pvar->pkt_state.datastart = 0;
  pvar->pkt_state.datalen = 0;
  pvar->pkt_state.seen_server_ID = FALSE;
  pvar->pkt_state.seen_newline = FALSE;
  pvar->pkt_state.predecrypted_packet = FALSE;
}

/* Read some data, leave no more than up_to_amount bytes in the buffer,
   return the number of bytes read or -1 on error or blocking. */
static int recv_data(PTInstVar pvar, unsigned long up_to_amount) {
  int amount_read;

  /* Shuffle data to the start of the buffer */
  if (pvar->pkt_state.datastart != 0) {
    memmove(pvar->pkt_state.buf,
      pvar->pkt_state.buf + pvar->pkt_state.datastart,
      pvar->pkt_state.datalen);
    pvar->pkt_state.datastart = 0;
  }

  buf_ensure_size(&pvar->pkt_state.buf, &pvar->pkt_state.buflen, up_to_amount);

  amount_read = (pvar->Precv)(
    pvar->socket,
    pvar->pkt_state.buf + pvar->pkt_state.datalen,
    up_to_amount - pvar->pkt_state.datalen, 0);

  if (amount_read > 0) {
    /* Update seen_newline if necessary */
    if (!pvar->pkt_state.seen_server_ID && !pvar->pkt_state.seen_newline) {
      int i;

      for (i = 0; i < amount_read; i++) {
        if (pvar->pkt_state.buf[pvar->pkt_state.datalen + i] == '\n') {
          pvar->pkt_state.seen_newline = 1;
        }
      }
    }

    pvar->pkt_state.datalen += amount_read;
  }

  return amount_read;
}

/* This function does two things:
   -- reads data from the sshd and feeds the SSH protocol packets to ssh.c
   -- copies any available decrypted session data into the application buffer
*/
int PKT_recv(PTInstVar pvar, char FAR * buf, int buflen) {
  int amount_in_buf = 0;
  BOOL connection_closed = FALSE;

  while (SSH_is_any_payload(pvar) ? buflen > 0 : !connection_closed) {
    if (SSH_is_any_payload(pvar)) {
      /* ssh.c has some session data for us to give to Teraterm. */
      int grabbed = SSH_extract_payload(pvar, buf, buflen);

      amount_in_buf += grabbed;
      buf += grabbed;
      buflen -= grabbed;
    } else if (!pvar->pkt_state.seen_server_ID &&
      (pvar->pkt_state.seen_newline || pvar->pkt_state.datalen >= 255)) {
      /* We're looking for the initial ID string and either we've seen the
         terminating newline, or we've exceeded the limit at which we should see
         a newline. */
      unsigned int i;

      for (i = 0; pvar->pkt_state.buf[i] != '\n' && i < pvar->pkt_state.datalen; i++) {
      }
      if (pvar->pkt_state.buf[i] == '\n') {
        i++;
      }

      if (SSH_handle_server_ID(pvar, pvar->pkt_state.buf, i)) {
        pvar->pkt_state.seen_server_ID = 1;
      }

      pvar->pkt_state.datastart += i;
      pvar->pkt_state.datalen -= i;
    } else if (pvar->pkt_state.seen_server_ID
      && pvar->pkt_state.datalen >= (unsigned int)SSH_get_min_packet_size(pvar)) {
      char FAR * data = pvar->pkt_state.buf + pvar->pkt_state.datastart;
      uint32 padding;
      uint32 pktsize;
      uint32 total_packet_size;

      if (!pvar->pkt_state.predecrypted_packet) {
        SSH_predecrpyt_packet(pvar, data);
        pvar->pkt_state.predecrypted_packet = TRUE;
      }

      if (SSHv1(pvar)) {
        uint32 realpktsize = get_uint32_MSBfirst(data);

        padding = 8 - (realpktsize % 8);
        pktsize = realpktsize + padding;
      } else {
        pktsize = get_uint32_MSBfirst(data);
        padding = (unsigned char)data[4];
      }

      total_packet_size = pktsize + 4 + SSH_get_clear_MAC_size(pvar);
      
      if (total_packet_size <= pvar->pkt_state.datalen) {
        /* the data must be 4 byte aligned. */
        SSH_handle_packet(pvar, data, pktsize, padding);
        pvar->pkt_state.predecrypted_packet = FALSE;

        pvar->pkt_state.datastart += total_packet_size;
        pvar->pkt_state.datalen -= total_packet_size;
      } else if (total_packet_size > 4*1024*1024) {
        notify_fatal_error(pvar, "Oversized packet received from server; connection will close.");
      } else {
        int amount_read = recv_data(pvar, max(total_packet_size, READAMOUNT));
        
        if (amount_read == SOCKET_ERROR) {
          if (amount_in_buf == 0) {
            return SOCKET_ERROR;
          } else {
            return amount_in_buf;
          }
        } else {
          if (amount_read == 0) {
            connection_closed = TRUE;
          }
        }       
      }
    } else {
      int amount_read = recv_data(pvar, READAMOUNT);
      
      if (amount_read == SOCKET_ERROR) {
        if (amount_in_buf == 0) {
          return SOCKET_ERROR;
        } else {
          return amount_in_buf;
        }
      } else if (amount_read == 0) {
        connection_closed = TRUE;
      }
    }

    if (pvar->fatal_error) {
      return amount_in_buf;
    }
  }

  if (SSH_is_any_payload(pvar)) {
    PostMessage(pvar->NotificationWindow, WM_USER_COMMNOTIFY,
      pvar->socket, MAKELPARAM(FD_READ, 0));
  }
  
  return amount_in_buf;
}

void PKT_end(PTInstVar pvar) {
  buf_destroy(&pvar->pkt_state.buf, &pvar->pkt_state.buflen);
}
